from .spectral2 import get_default_spectral2_object, calculate_spectrum_spectral2
from .smarts import get_default_smarts_object, calculate_spectrum_smarts, SmartsSolverError
from .light_source import reference_spectra, LightSource



